/*
  Copyright (c) 2004 Dirk Jagdmann <doj@cubic.org>

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any
  purpose, including commercial applications, and to alter it and
  redistribute it freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must
     not claim that you wrote the original software. If you use this
     software in a product, an acknowledgment in the product
     documentation would be appreciated but is not required.

  2. Altered source versions must be plainly marked as such, and must
     not be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source
     distribution.
*/

#ifdef _WIN32
#pragma warning( disable : 4786 )
#endif

#include <iostream>
#include <map>
#include <string>
using namespace std;

#ifdef __linux__
#include <cstring>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#endif

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

namespace doj
{

#ifdef __linux__
  /**
     data needed to handle a memory map on linux systems.
  */
  struct mmap_info {
    int fh;
    int filelen;
    string filename;
  };
#endif

#ifdef _WIN32
  /**
     data needed to handle a memory map on WIN32 systems.
  */
  struct mmap_info {
    HANDLE file;
    HANDLE map;
    int filelen;
    string filename;
  };
#endif

  typedef map<void*, struct mmap_info> registry_t;

  /**
     the memorymap registry contains all active memory maps and the
     needed data. This information is needed for unmapping and to
     retrieve the map size via the memorymapsize(...) function.
  */
  static registry_t registry;

#ifdef __linux__

  void* memorymap(const char* filename, const bool readonly)
  {
    struct mmap_info info;
    info.filename=filename;

    // get file size
    struct stat buf;
    if(stat(filename, &buf) < 0)
      {
#ifdef DOJDEBUG
	cerr << "memorymap(): could not fstat " << filename << " : " << strerror(errno) << endl;
#endif
	return 0;
      }
    if(!S_ISREG(buf.st_mode))
      {
#ifdef DOJDEBUG
	cerr << "memorymap(): " << filename << " is not a regular file" << endl;
#endif
	return 0;
      }
    info.filelen=buf.st_size;

    // open file
    info.fh=::open(filename, O_RDONLY);
    if(info.fh<0)
      {
#ifdef DOJDEBUG
	cerr << "memorymap(): could not open " << filename << " : " << strerror(errno) <<  endl;
#endif
	return 0;
      }

    // memory map file
    void *w=mmap(NULL, info.filelen, PROT_READ, MAP_PRIVATE, info.fh, 0);
    if(w==0)
      {
#ifdef DOJDEBUG
	cerr << "memorymap(): could not mmap " << filename << " : " << strerror(errno) <<  endl;
#endif
	::close(info.fh);
	return 0;
      }

    // make an entry into registry
    registry[w]=info;

    return w;
  }

  int memoryunmap(void *mem)
  {
    if(registry.count(mem)!=0)
      {
	struct mmap_info info=registry[mem];
	registry.erase(mem);
	if(munmap(mem, info.filelen) < 0)
	  {
#ifdef DOJDEBUG
	    cerr << "memoryunmap(): could not unmap " << info.filename << " : " << strerror(errno) << endl;
#endif
	  }
	if(::close(info.fh) < 0)
	  {
#ifdef DOJDEBUG
	    cerr << "memoryunmap(): could not close " << info.filename << " : " << strerror(errno) << endl;
#endif
	  }
      }
    else
      {
#ifdef DOJDEBUG
	cerr << "memoryunmap(): pointer not found in registry" << endl;
#endif
	return -1;
      }

    return 0;
  }
#endif // __linux__

#if _WIN32
  void* memorymap(const char* filename, const bool readonly)
  {
    struct mmap_info info;
    info.filename=filename;

    info.file=CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if(info.file==INVALID_HANDLE_VALUE)
      {
#ifdef DOJDEBUG
	cerr << " could not open " << GetLastError() << endl;
#endif
	return 0;
      }

    info.map=CreateFileMapping(info.file, NULL, PAGE_READONLY, 0, 0, filename);
    if(info.map==INVALID_HANDLE_VALUE)
      {
#ifdef DOJDEBUG
	cerr << " could not create map " << GetLastError() << endl;
#endif
	CloseHandle(info.file);
	return 0;
      }

    void *mem=MapViewOfFile(info.map, FILE_MAP_READ, 0, 0, 0);
    if(mem==NULL)
      {
#ifdef DOJDEBUG
	cerr << " could not map file " << GetLastError() << endl;
#endif
	CloseHandle(info.map);
	CloseHandle(info.file);
	return 0;
      }

    registry[mem]=info;
    return mem;
  }

  int memoryunmap(void *mem)
  {
    if(registry.count(mem)!=0)
      {
	struct mmap_info info=registry[mem];
	registry.erase(mem);
	UnmapViewOfFile(mem);
	CloseHandle(info.map);
	CloseHandle(info.file);
      }
    else
      {
#ifdef DOJDEBUG
	cerr << "memoryunmap(): pointer not found in registry" << endl;
#endif
	return -1;
      }

    return 0;
  }
#endif // _WIN32

  int memorymapsize(void *mem)
  {
    if(registry.count(mem)!=0)
      return registry[mem].filelen;
    return -1;
  }

};
